# perl

# Fix the mess that Dropbox made of our TLIB Version Control library files
# by David A. Burton, Burton Systems Software
# Uncopyrighted. 2016

# TLIB Version Control fills in the version information for us:
$version_str = "";
#--=>keyflag<=-- "&(#)%n, version %v, %d "
$version_str = "&(#)tlib_dropbox_fixer.pl, version 8, 10-Aug-16 ";

$! = 1;  # immediate output

sub do_help {
   print "perl tlib_dropbox_fixer.pl [options] [folder]\n" .
     "\n" .
     "Dropbox burned me!  It renamed several dozen current TLIB libraries with\n" .
     '"conflicted copy" in the name (so, for example, TLIB library abc.pl____$ became' . "\n" .
     '"abc (OWNER-PC\'s conflicted copy 2016-08-02).pl____$"), and replaced them with' . "\n" .
     "old versions: shorter, and with older file dates, with no warning whatsoever.\n" .
     "\n" .
     "To fix the mess, you need to compare the current version of each file with the\n" .
     '"conflicted copy" version, and verify that the longer version is the same as' . "\n" .
     "the shorter version, except with additional data appended, and then make the\n" .
     "longer version the current version. So that's what this program does.\n" .
     "\n" .
     "Run it with no parameters to see this Help message.\n" .
     "\n" .
     "Run it specifying the folder to be scanned (or just \".\" for current folder) to\n" .
     "see a \"report\" (of .bat file commmands) showing which files need to be renamed.\n" .
     "\n" .
     "Redirect output to a .bat file to create a script to do all the renames.\n" .
     "\n" .
     "Options:\n" .
     "  -r  to Recursively traverse subdirectories.\n" .
     "  -d  Debug (display verbose messages).\n" .
     "\n" .
     "Example:\n" .
     '  perl tlib_dropbox_fixer.pl -r C:\Users\Owner\Dropbox\tlib_libs\\ >fixem.bat' . "\n" .
     "  fixem.bat\n" .
     "\n" .
     "Note: This program is compatible with both Perl 4 and Perl 5 for Windows.\n";
   exit(1);
}

$rd_size = 16 * 1024 * 1024;  # set buffer size for file reads to 16 MB
$slash = '\\';  # slash directory-separator used by the file system (default is a backslash for Windows)
$rc = 0;

$target_folder = '';  # where are the files we should fix?
$recurse = 0;    # -r  by default, do not go recursive
$debugmode = 0;  # -d  by default do not be verbose

# handle the command line parameters
while ($#ARGV >= 0) {
   $parm = shift(@ARGV);
   if ($parm =~ /^\-/) {
      if ($parm eq '-r') {
         $recurse = 1;
      } elsif ($parm eq '-d') {
         $debugmode = 1;
      } else {
         print STDERR "ERROR: unrecognized option: '$parm'\n";
         exit(1);
      }
   } else {
      if ('' ne $target_folder) {
         print STDERR "ERROR: you may only specify one target folder.\n";
         exit(1);
      }
      $target_folder = $parm;
   }
}

# if they didn't specify a target folder, display the Help message
if ('' eq $target_folder) {
   &do_help;
   exit(1);
}

# target folder has to end in a slash (forward or backward), so if the slash is missing then add it
$lastch = substr($target_folder,-1,1);
if (($lastch ne '/') && ($lastch ne '\\')) {
   $target_folder .= $slash;
} else {
   # if user specified a trailing slash, then use that kind of slash
   $slash = $lastch;
}

$rc = 1;

# side-effect result of cmp2files()
$flen1 = $flen2 = 0;

# Binary file compare.
# Input is two file names.
# Result is an integer:
#   0 = the files are identical.
#   1 = file1 is superset of file2 ("file1 is best"). File2 is identical to the first part of file1.
#   2 = file2 is superset of file1 ("file2 is best"). E.g., a newer version of a TLIB library or appending log file.
#   3 = the files are unequal, and neither is a superset of the other.
#   4 = error: e.g., one or both files were not found.
# Error messages are written top STDERR, but otherwise this subroutine is silent.
sub cmp2files {
   local($file1,$file2) = @_;
   local($rc) = 0;  # result
   local($rc1,$rc2,$tmp,$buf1,$buf2,$firstloop);
   $flen1 = $flen2 = 0; # side-effect result
   if (!defined $rd_size) {
      $rd_size = 1024 * 1024;  # default buffer size for file reads is 1 MB
   }
   # print STDERR "cmp2files('$file1','$file2')...\n";
   if (!open(FD1, '<'.$file1)) {
      # note: prepending the '<' avoids accidental output if someone passes in a file name that begins with '>'
      print STDERR "$!, \"$file1\"\n";
      $rc = 4;  # file not found or could not be opened
      return $rc;
   }
   binmode FD1;
   if (!open(FD2, '<'.$file2)) {
      print STDERR "$!, \"$file2\"\n";
      $rc = 4;  # file not found or could not be opened
      close FD1;
      return $rc;
   }
   binmode FD2;
   $buf1 = $buf2 = '';
   $firstloop = 1;
   for (;;) {
      $rc1 = sysread(FD1, $buf1, $rd_size);
      $rc2 = sysread(FD2, $buf2, $rd_size);
      # print STDERR "read $rc1 bytes from '$file1' and $rc2 bytes from '$file2'\n";
      if (!defined $rc1) {
         print STDERR "$!\n";
         $rc = 4;
         last;
      }
      if (!defined $rc2) {
         print STDERR "$!\n";
         $rc = 4;
         last;
      }
      $flen1 += $rc1;
      $flen2 += $rc2;
      if ($rc1 != length($buf1)) {
         $tmp = length($buf1);
         print STDERR "ERR: read of $rc1 bytes from \"$file1\" returned $tmp-byte buffer!\n";
      }
      if ($rc2 != length($buf2)) {
         $tmp = length($buf2);
         print STDERR "ERR: read of $rc2 bytes from \"$file2\" returned $tmp-byte buffer!\n";
      }
      if ($rc1 != $rc2) {
         # files are different lengths, but one might be a superset of the other
         $rc = 3;
         if ($rc1 > $rc2) {
            $tmp = substr($buf1,0,length($buf2));
            if ($tmp eq $buf2) {
               # file1 is a superset of file2
               $rc = 1;
            }
         } else {
            $tmp = substr($buf2,0,length($buf1));
            if ($tmp eq $buf1) {
               # file2 is a superset of file1
               $rc = 2;
            }
         }
         last;
      }
      if ($buf1 ne $buf2) {
         # the amount read so far from the two files is the same, but the file contents are different
         $rc = 3;
         last;
      }
      if (0 == $rc1) {
         # we're at EOF for both files, so the files are identical
         $rc = 0;
         # ...unless they're 0-length, in which case something went wrong!
         if ((0==$flen1) && $firstloop) {
            print STDERR "ERROR: input files appear to be zero-length, $!\n";
            $rc = 4;
         }
         last;
      }
      $firstloop = 0;
   }#for
   close FD1;
   close FD2;
   # print STDERR "cmp2files('$file1','$file2') done, rc=$rc.\n";
   return $rc;
}#cmp2files


# $rc = &cmp2files("./file1", "./file2");  # test cmp2files()
# print "cmp2files('./file1', './file2') = $rc\n";


$file_fix_cntr = 1;  # used when generating GOTO line labels


# Find all the "conflicted copy" files in a folder, compare them with the
# original names, and do the necessary renames.
# If $recurse in 1..19 then recursively process subdirectories, too.
# ($recurse also counts the recursion depth)
sub process_one_directory {
   local($srchdir, $recurse) = @_;  # note: it is assumed that $dirpath ends in a slash
   local(@subdirs) = ();
   local($sdir, $fn, $fn2, $fn3);

   # read the directory
   $sdir = $srchdir;
   if ((length($srchdir) > 1) && (':' ne substr($srchdir,-2,1))) {
      # Except for '\' or 'D:\' remove the trailing slash (necessary for Perl 4, harmless for Perl 5)
      $sdir = substr($srchdir,0,length($srchdir)-1);
   }
   if ($debugmode) {
      print STDERR "dbg: srchdir='$srchdir', sdir='$sdir'\n";
   }
   opendir(SRCHDIR,$sdir) || die "Could not read directory: \"$sdir\", $!\n";
   local(@allfiles) = readdir(SRCHDIR);
   closedir(SRCHDIR);
   if ($debugmode) {
      printf STDERR "Found %d files & folders in \"%s\"\n\n", 1+$#allfiles, $srchdir;
   }

   # Look for files with " conflicted copy " in the name; if we find one, then
   # deduce the original name, compare the two, and decide if the "conflicted copy"
   # is better.
   for $fn (@allfiles) {
      if (("." eq $fn) || (".." eq $fn)) {
         next;  # ignore . and .. folders
      }
      if (-d "$srchdir$fn") {
         # print STDERR "dbg: DIR $srchdir$fn\n";
         push( @subdirs, "$srchdir$fn$slash" );
      } else {
         # print STDERR "dbg:    $srchdir$fn\n";
         if ($fn =~ / conflicted copy /) {
            # Got one; now figure out the original file name
            $fn2 = $fn;
            $fn2 =~ s/ \([^\(]* conflicted copy [0-9\-]+\)//;
            # $fn2 should be the original file name
            if ($fn eq $fn2) {
               print STDERR "Warning: could not deduce original file name from \"$fn\"\n";
            } elsif (!-f "$srchdir$fn2") {
               print STDERR "Warning: found \"$fn\" but not \"$fn2\" in \"$srchdir\"\n";
            } else  {
               if ($debugmode) {
                  print STDERR "Comparing \"$srchdir$fn2\" to \"$srchdir$fn\"...\n";
               }
               $rc = &cmp2files("$srchdir$fn","$srchdir$fn2");
               # print STDERR "rc=$rc\n";
               if ($rc) {
                  if (1==$rc) {
                     if ($debugmode) {
                        print "\"$srchdir$fn\" is a superset of \"$srchdir$fn2\"\n";
                     }
                     print "\n" .
                           "REM -- restore newer/better \"conflicted copy\" of TLIB library file \"$fn2\"\n" .
                           "attrib -r \"$srchdir$fn\"\n" .
                           "attrib -r \"$srchdir$fn2\"\n" .
                           "if not exist \"$srchdir$fn\" goto L$file_fix_cntr\n" .  # avoid catastrophe if someone runs the .bat file twice!
                           "del \"$srchdir$fn2\"\n" .
                           "ren \"$srchdir$fn\" \"$fn2\"\n" .
                           ":L$file_fix_cntr\n";
                  } elsif (2==$rc) {
                     if ($debugmode) {
                        print "\"$srchdir$fn2\" is a superset of \"$srchdir$fn\"\n";
                     }
                     $fn3 = $fn;
                     $fn3 =~ s/ conflicted / obsolete /;
                     print "\n" .
                           "REM -- the \"conflicted copy\" of TLIB library file \"$fn2\" is out-of-date\n" .
                           "attrib -r \"$srchdir$fn\"\n" .
                           "ren \"$srchdir$fn\" \"$fn3\"\n";
                  } elsif (3==$rc) {
                     print "\n" .
                           "REM -- can't tell what to do with \"$srchdir$fn\"\n" .
                           "echo WARNING: \"$srchdir$fn2\" is\n" .
                           "echo   unequal to \"$srchdir$fn\"\n" .
                           "echo   and neither is a truncated version of the other.\n" .
                           "pause\n";
                  } else {
                     print STDERR "ERROR rc=$rc.\n";
                  }
               } else {
                  if ($debugmode) {
                     print STDERR "files are identical, rc=0.\n";
                  }
                  if ($debugmode) {
                     print "\"$srchdir$fn2\" is identical to \"$srchdir$fn\"\n";
                  }
                  $fn3 = $fn;
                  $fn3 =~ s/ conflicted / duplicate /;
                  print "\n" .
                        "REM -- the \"conflicted copy\" is a duplicate of TLIB library file \"$fn2\"\n" .
                        "attrib -r \"$srchdir$fn\"\n" .
                        "ren \"$srchdir$fn\" \"$fn3\"\n";
               }
            }
            if ($debugmode) {
               print STDERR "\n";
            }
            $file_fix_cntr++;
         }
      }
   }#for
   if ($recurse && ($recurse < 20)) {  # limit recursion depth to 20 -- should never happen
      for $sdir (@subdirs) {
         &process_one_directory($sdir,1+$recurse);
      }
   }
}#process_one_directory


# Main program: do it!
&process_one_directory($target_folder,$recurse);

exit(0);

